from typing import Union
from math import log

from ..others import Humidity
from ..base.Decorators import UnitType
from ..base import Measurement, Dimension


@UnitType
class Temperature(Measurement, metaclass=Dimension, symbol='Θ'):
	Celsius: type
	Fahrenheit: type
	Kelvin: type

	_decorator = 'º'
	_id = 'ºt'

	def __new__(cls, value: float | int | Measurement):
		if isinstance(value, Temperature) and not isinstance(value, cls):
			value = value[cls.unit]
		return super().__new__(cls, value)

	@property
	def celsius(self):
		from ..temperature import Celsius
		return Celsius(self._celsius())

	@property
	def fahrenheit(self):
		from ..temperature import Fahrenheit
		return Fahrenheit(self._fahrenheit())

	@property
	def kelvin(self):
		from ..temperature import Kelvin
		return Kelvin(self._kelvin())

	@property
	def fDelta(self):
		from ..temperature import Fahrenheit
		return Fahrenheit(self._fahrenheit(delta=True))

	@property
	def cDelta(self):
		from ..temperature import Celsius
		return Celsius(self._celsius(delta=True))

	def dewpoint(self, rh: float):
		from ..temperature import Celsius

		# Keep rh within reasonable ranges
		rh = self.normalizeRh(rh)

		# a, b = 17.27, 237.7
		# n = ((a * T) / (b + T)) + logRH
		# value = Celsius((b * n) / (a - n))

		T = float(self.c)
		logRH = log(rh/100)
		A = 243.04
		B = 17.625
		value = Celsius(A*(logRH + ((B*T)/(A + T)))/(B - logRH - ((B*T)/(A + T))))

		value = self.__class__(value[self._unit])
		value = self.transform(value)
		value.title = 'Dewpoint'
		value.calculated = True
		return self.transform(value)

	def heatIndex(self, rh: float):
		R = float(self.normalizeRh(rh))

		if 300 > self.kelvin < 318 or R < 13:
			return self

		if self._unit == 'f':
			c = [-42.379, 2.04901523, 10.14333427, -0.22477541, -0.00683783, -0.05481717, 0.00122874, 0.00085282, -0.00000199]
		else:
			c = [-8.78469475556, 1.61139411, 2.33854883889, -0.14611605, -0.012308094, -0.0164248277778, 0.002211732, 0.00072546, -0.000003582]

		T = float(self if self._unit != 'k' else self.c)
		T2 = pow(T, 2)
		R2 = pow(R, 2)

		hi = c[0] + (c[1]*T) + (c[2]*R) + (c[3]*T*R) + (c[4]*T2) + (c[5]*R2) + (c[6]*T2*R) + (c[7]*T*R2) + (c[8]*T2*R2)

		hi = hi if self._unit != 'k' else hi + 273.15
		value = self.transform(self.__class__(hi))
		value.title = 'Heat Index'
		value.calculated = True
		return value

	def windChill(self, wind: 'Wind'):
		if wind.mph < 3:
			return type(self)(self)
		if self._unit == 'f':
			w = float(wind.mih)
			t = float(self)
			value = 35.74 + (0.6215*t) - (35.75*pow(w, 0.16)) + (0.4275*t*pow(w, 0.16))
		else:
			w = float(wind.kmh)
			t = float(self.c)
			value = 13.12 + (0.6215*t) - (11.37*pow(w, 0.16)) + (0.3965*t*pow(w, 0.16))

		value = self.transform(self.__class__(round(value, self._precision)))
		value.title = 'Wind Chill'
		value.calculated = True
		return value

	@staticmethod
	def normalizeRh(rh: Union[int, float, 'Humidity']):
		if not isinstance(rh, Humidity):
			rh = Humidity(rh)
		return int(rh)

	def _convert(self, value: Measurement):
		if isinstance(value, Measurement):
			return value[self.unit]
		return value[self.unit]

	# abbreviations
	c = celsius
	k = kelvin
	kel = kelvin
	f = fahrenheit





__all__ = ['UnitType', 'Synonym', 'Tiny', 'Small', 'Medium', 'Large', 'Huge']


def UnitType(*args, **kwargs):
	if args:
		cls = args[0]
		cls._type = cls
		return cls
	else:
		def wrapper(cls):
			cls._type = cls
			return cls

		return wrapper


def strToDict(string: str, cls: type) -> type:

	def parseString(item: str):
		key, value = item.split('=')
		# expectedTypes = {'max': int, 'precision': int, 'unitSpacer': stringToBool, 'shorten': stringToBool, 'thousandsSeparator': stringToBool, 'cardinal': stringToBool, 'degrees': stringToBool}
		if value.isnumeric():
			value = float(value)
			if value.is_integer():
				value = int(value)
		if value == 'True':
			value = True
		elif value == 'False':
			value = False
		return f'_{key}', value

	conf = [parseString(a) for a in [(y.strip(' ')) for y in string.split(',')]]
	for item in conf:
		setattr(cls, *item)

	return cls


def Tiny(cls):
	cls._size = 'tiny'
	return cls


def Small(cls):
	cls._size = 'small'
	return cls


def Medium(cls):
	cls._size = 'medium'
	return cls


def Large(cls):
	cls._size = 'large'
	return cls


def Huge(cls):
	cls._size = 'huge'
	return cls


def Synonym(cls):
	cls.__name__ = cls.__mro__[1].__name__
	return cls



import logging
import re
from builtins import float, isinstance
from collections import ChainMap, namedtuple
from difflib import get_close_matches
from functools import lru_cache
from locale import delocalize
from typing import ClassVar, Optional, Set, Type, Union, Tuple, ForwardRef, TypeVar, Literal, Final, Mapping, Iterable
from math import nan, isnan, inf
from decimal import Decimal

from ..utils import modifyCase, pluralize, empty, getFrom, UnsetKwarg, loadUnitLocalization, CaseInsensitiveKey
from ..config import config, GROUPING_CHAR, RADIX_CHAR

__all__ = ('SmartFloat', 'Limits', 'TypedLimits', 'FormatSpec', 'FiniteField', 'UnitDict', 'MetaUnitClass')

regexType: Final = Literal['b', 'c', 'd', 'e', 'E', 'f', 'F', 'g', 'G', 'n', 'o', 's', 'x', 'X', '%']

log = logging.getLogger('SmartFloat')

__all__ = ['SmartFloat', 'FiniteField', 'MetaUnitClass', 'FormatSpec']

Measurement = ForwardRef('Measurement', is_class=True, module='Measurement')
_T = TypeVar('_T', Measurement, float)

Limits = namedtuple('Limits', 'min max')


class TypedLimits(Limits):
	min: _T
	max: _T

	@classmethod
	def fromLimits(cls, limit: tuple, type: _T = float) -> 'TypedLimits':
		return cls(type(limit[0]), type(limit[1]))

	def cast(self, type: _T) -> 'TypedLimits':
		return self.fromLimits(self, type)


class FormatSpec:
	truthy_values: ClassVar[Set[str]] = {'true', 't', 'yes', 'y', 'on', 'shown', 'show'}
	falsy_values: ClassVar[Set[str]] = {'false', 'f', 'no', 'n', 'off', 'hidden', 'hide'}
	limit = re.compile(r'\[(?P<max>([+-]?[\d.]+)|\*)?:(?P<min>([+-]?[\d.]+)|\*)?]')
	precision = re.compile("""	
	^:?
	(?P<format_spec>
		(?P<align>(?P<fill>.)?[<^>])?
		(?P<sign>[-+\s])?
		(?P<alt>\#)?
		(?P<minwidth>\d+)?
		(?P<grouping_option>[_,])?
		(?P<radix>\.)?
		(?(radix)(?P<precision>\d+)|)?
		(?P<type>[bcdeEfFgGnosxX%])?
	)$
	""", re.VERBOSE)
	params = re.compile("""
	(
    (?P<keyquote>[\'\"`]?)    # optional start quote
    (?P<key>\S+?)
    (?P=keyquote)
	)
	:\s*
	(
    (?P<valquote>[\'\"`]?)     # optional start quote
    (?P<value>.*?)           # literal value
    (?P=valquote)
	)
	(,|$)
	""", re.VERBOSE)
	formatParams = re.compile("""{(?P<name>\w+?) (?: :(?P<spec>.*?))?}""", re.VERBOSE)
	conversion = re.compile(r"""
	^(?P<fullmatch>
		(?P<convertTo>[-\w\s/\\]+)
		(?P<delinator>:{1,2})?)
	(?P<rest>
		(?(delinator) # Test for delinator 
			(?: # if True
				# Option 1: [:][any
				(?:[-\w\s/\\]+?:).*$
				| # or 
				# Option 2: [::][format_spec][$]
				(?:\S[^:]+?$)
			)
		| # else
			$ # False
		)
	)
	""", re.VERBOSE)
	number = re.compile(fr"""
	(?P<number>
	[-+]?
	([\d{GROUPING_CHAR}]+)?
	([{RADIX_CHAR}]\d+)?)
	""", re.VERBOSE)
	true = re.compile(rf'^({"|".join(truthy_values)})$', re.IGNORECASE)
	false = re.compile(rf'^({"|".join(falsy_values)})$', re.IGNORECASE)

	valueParams = re.compile(rf'(value|self):{precision.pattern}', re.VERBOSE)

	@classmethod
	def _float(cls, s: str | float | int) -> Optional[float | int]:
		try:
			value = float(delocalize(str(s)))
			return int(value) if value.is_integer() else value
		except ValueError:
			return None

	@classmethod
	def getNumber(cls, value: str, strict: bool = True, default: Optional[float] = None) -> Optional[float]:
		"""
		Get a number from a string.
		:param value: string to get number from:
		:param strict: Only return value if the string is entirely numeric:
		:return: float or None
		"""

		if strict:
			value = cls._float(value)
			return value if value is not None else default
		results = next((i.groupdict() for i in cls.number.search(value) if i is not None or i != ''), None)
		return cls._float(results['number']) if results else default

	@classmethod
	def getBool(cls, value: str, strict: bool | None = True, default: Optional[bool] = None) -> Optional[bool]:
		"""
		Get a boolean from a string.
		:param value: string to get boolean from
		:param strict: Only return value if the string is entirely boolean
		:param default: Default value if the string is not boolean
		:return: bool or None
		"""
		value = value.lower()

		if strict:
			isTrue = cls.true.match(value)
			isFalse = cls.false.match(value)

			if isTrue == isFalse:
				return default
			if isTrue is not None:
				return True
			elif isFalse is not None:
				return False
			else:
				return default

		elif strict is None:
			hasTrue = value in cls.truthy_values
			hasFalse = value in cls.falsy_values
		else:
			hasTrue = any(i == value for i in cls.truthy_values)
			hasFalse = any(i == value for i in cls.falsy_values)

		if hasTrue == hasFalse:
			return default
		return hasTrue


class FiniteField:
	_limits: Limits

	def _limitFunc(self, value: float) -> float:
		return value%self._limits[1]


class UnitDict(dict):

	def __getitem__(self, k):
		k = CaseInsensitiveKey(k)
		return super().__getitem__(k)

	def __setitem__(self, k, v):
		k = CaseInsensitiveKey(k)
		super().__setitem__(k, v)

	def __delitem__(self, k) -> None:
		k = CaseInsensitiveKey(k)
		super().__delitem__(k)


class MetaUnitClass(type):
	_name: ClassVar[Optional[str]]
	_derived: ClassVar[bool]
	_unit: ClassVar[Optional[str]] = None
	_limits: ClassVar[Tuple[Union[int, float]]]
	_acceptedTypes: ClassVar[Set[Type]]

	class __Dimensionless__:
		_dimension = None
		_unit = None
		_convertable = False

		@property
		def dimension(self):
			return None

		@property
		def unit(self):
			return ''

		@property
		def convertable(self):
			return False

		def _convert(self, *args, **kwargs):
			return self

	class __NonPlural__:

		def __subclasscheck__(self, subclass):
			return issubclass(subclass, type(self))

		@property
		def pluralName(self) -> None:
			return None

		@property
		def pluralUnit(self) -> None:
			return None

	def __new__(mcs, name, bases, attrs, **kwargs):
		if kwargs.get('baseUnit', None):
			attrs['baseUnit'] = kwargs['baseUnit']

		# find the Measurement name
		measurementName = getFrom(
			('fullName', 'name', 'full_name', 'unitName', 'unit_name', 'unitFullName', 'unit_full_name'),
			kwargs, attrs,
			default=None, pop=True, expectedType=str
		) or modifyCase(name, joiner=' ')

		if measurementName:
			attrs['_name'] = measurementName

		if aliases := getFrom(
			('aliases', 'alias', 'alternate'),
			kwargs, attrs,
			default=None, pop=True, findAll=True, expectedType=Iterable
		):
			if isinstance(aliases, str):
				aliases = (aliases,)
			if aliases and isinstance(aliases, list) and not isinstance(aliases[0], str):
				aliases = tuple(aliases[0])
			attrs['_aliases'] = set(aliases)

		# find the unit if not dimensionless
		if not any([issubclass(base, mcs.__Dimensionless__) for base in bases]):
			if (unit := getFrom(
				('_unit', 'unit'),
				kwargs, attrs,
				default=None, pop=True, expectedType=str)
			) is not None:
				attrs['_unit'] = unit

		# find the plurals if not non-plural
		if not any([issubclass(base, mcs.__NonPlural__) for base in bases]):
			if unit := attrs.get('_unit', None):
				attrs['_pluralUnit'] = getFrom(
					('pluralUnit', 'plural_unit', 'plural'),
					kwargs, attrs,
					default=None, pop=True, expectedType=str
				) or pluralize(unit)

			attrs['_pluralName'] = getFrom(
				('pluralName', 'plural_name', 'pluralFullName', 'plural_full_name'),
				kwargs, attrs,
				default=None, pop=True, expectedType=str
			) or pluralize(measurementName)

		# set convertable if found in kwargs
		if (convertable := kwargs.get('convertable', None)) is not None:
			attrs['_convertable'] = convertable

		if limits := getFrom(
			('limits', 'limit', 'minMax', 'min_max'),
			kwargs, attrs,
			default=None, pop=True, expectedType=Iterable,
			findAll=False,
		):
			if None in limits:
				m = limits[0]
				m = m if m is not None else min(getattr(mcs, '_limits', (-inf, inf)))
				M = limits[1]
				M = M if M is not None else max(getattr(mcs, '_limits', (-inf, inf)))
				limits = sorted((m, M))
			if isinstance(limits, dict):
				_max = limits.get('max', None) or max(getattr(mcs, '_limits', (-inf, inf)))
				_min = limits.get('min', None) or min(getattr(mcs, '_limits', (-inf, inf)))
				limits = sorted(_min, _max)
			attrs['_limits'] = Limits(*limits)

		# load attrs from config
		if name == 'Measurement':
			global Measurement
			attrs.update({key: value for key, value in config.unitDefaults.items() if value is not None})
			# attrs['__unitDict__'] = ChainMap()
			Measurement = super().__new__(mcs, name, bases, attrs, **kwargs)
			return Measurement
		# unitDict = ChainMap()
		# for base in genericBases:
		# 	base.__unitDict__.maps.append(unitDict)
		# attrs['__unitDict__'] = unitDict

		if name.lower() in config.unitPropertiesKeys:
			configPropertiesKey = get_close_matches(name, list(config.unitProperties.keys()), cutoff=0.3)
			configProps = {}
			for item in config.unitProperties[configPropertiesKey[0]].split(','):
				key, value = item.strip(' ').split('=')
				if (number := FormatSpec.getNumber(value, strict=True)) is not None:
					value = number
				elif (boolValue := FormatSpec.getBool(value, strict=False)) is not None:
					value = boolValue if boolValue is not None else empty
				configProps[f'_{key}'] = value
			attrs.update(configProps)

		attrs['__annotations__'] = ChainMap(attrs.get('__annotations__', {}), *(b.__dict__.get('__annotations__', {}) for b in bases))

		return super().__new__(mcs, name, bases, attrs, **kwargs)

	@property
	def isGeneric(cls) -> bool:
		return (cls is Measurement and cls is not SmartFloat) or getattr(cls, '_type', None) is cls

	@property
	def genericBases(self):
		return {base for base in self.__mro__ if getattr(base, 'isGeneric', False) or getattr(base, '_system', None) is base}

	@property
	def Generic(self) -> Type:
		genericBases = self.genericBases
		return next((base for base in genericBases if base is not self), None)

	def __subclasscheck__(cls, subclass):
		if subclass is None:
			return False
		value = super().__subclasscheck__(subclass)
		return value

	@property
	@lru_cache(maxsize=128)
	def localizedUnit(self) -> Type[Measurement] | Tuple[Type[Measurement] | None, ...] | None:
		unit = self.__parse_unit__(loadUnitLocalization(self, config))
		if isinstance(unit, (tuple, list)):
			if '*' in unit:
				if unit[0] == '*':
					unit = self.numerator, self.__findUnitClass__(unit[1])
				if unit[1] == '*':
					unit = self.__findUnitClass__(unit[0]), self.denominator
				return unit
			else:
				return tuple(self.__findUnitClass__(i) for i in unit)
		return self.__findUnitClass__(unit)

	def __parse_unit__(self, unit: str) -> str:
		return unit

	@property
	def isDerived(cls) -> bool:
		if not hasattr(cls, '_derived'):
			cls._derived = getattr(cls.__parent_class__, 'isDerived', None)
		return cls._derived

	@property
	def isScaling(cls) -> bool:
		return hasattr(cls, '_Scale')

	@property
	def __parent_class__(cls) -> Measurement:
		return cls.__mro__[1]

	@property
	def type(cls) -> Type[Measurement]:
		return getattr(cls, '_type', None) or next((base for base in cls.__mro__ if getattr(base, 'isGeneric', False)), cls)

	@property
	def dimension(self):
		return getattr(self, '__dimension__', None)

	@property
	def subTypes(cls):
		subs: Set[Type['Measurement']] = set()
		for sub in cls.__subclasses__():
			if sub.dimension == cls.dimension:
				subs.add(sub)
			subs.update(getattr(sub, 'subTypes', set()))
		return subs

	@property
	def unit(cls) -> str:
		return cls._unit or ''

	@property
	def system(self):
		return getattr(self, '_system', None)

	@property
	def name(self):
		return getattr(self, '_name', self.__name__)

	@property
	def decorator(cls):
		return cls._decorator

	@property
	def suffix(cls):
		return getattr(cls, '_suffix', '')

	@property
	def id(cls):
		return (
			cls.__dict__.get('_id', None)
			or getattr(cls, '_unit', None)
			or getattr(cls, '_decorator', None)
			or getattr(cls, '_suffix', None)
			or getattr(cls, '_id', None)
			or getattr(cls, 'unit', None)
			or getattr(cls, 'name', None)
		)

	@property
	def compatibleUnits(cls) -> UnitDict:
		units = UnitDict()
		for subtype in cls.type.subTypes:
			units.update({alias.lower(): subtype for alias in (subtype.unit, *subtype.aliases) if alias})
		units.pop(None, None)
		return units

	@property
	def compatibleTypes(cls) -> Set[Measurement]:
		return cls._acceptedTypes

	@property
	def fixedUnits(cls) -> tuple[Optional[Type[Measurement]], Optional[Type[Measurement]]]:
		n: Type['Measurement'] = getattr(cls, '_numerator', None)
		d: Type['Measurement'] = getattr(cls, '_denominator', None)
		if n is not None:
			n = None if n.isGeneric else n
		if d is not None:
			d = None if d.isGeneric else d
		return n, d

	@property
	def limits(cls) -> Limits:
		if cls.isDerived:
			return getattr(cls, '_limits', None) or getattr(cls.numeratorClass, 'limits', None) or (-inf, inf)
		return cls._limits

	@property
	def typedLimits(cls) -> TypedLimits:
		return TypedLimits.fromLimits(cls.limits, type=cls)

	def __repr__(cls):
		if cls.isDerived:
			match cls.fixedUnits:
				case MetaUnitClass(), MetaUnitClass():
					if '[' in cls.__name__:
						return cls.__name__
					return f'{cls.__name__}[{cls.numerator.__name__}/{cls.denominator.__name__}]'
				case MetaUnitClass(), None:
					return f'{cls.__name__}[{cls.denominator.__name__}]'
				case None, MetaUnitClass():
					return f'{cls.__name__}[{cls.numerator.__name__}]'
				case _:
					pass

			if ((cls.numerator.isGeneric or cls.__parent_class__.numerator.isGeneric)
				and (cls.denominator.isGeneric or cls.__parent_class__.denominator.isGeneric)):
				return f'{cls.__name__}[{repr(cls.numerator)}/{repr(cls.denominator)}]'
			elif cls.numerator.isGeneric or cls.__parent_class__.numerator.isGeneric:
				return f'{cls.__name__}[{repr(cls.numeratorClass)}]'
			elif cls.denominator.isGeneric or cls.__parent_class__.denominator.isGeneric:
				return f'{cls.__name__}[{repr(cls.denominator)}]'
			else:
				return f'{cls.__name__}'
		if cls.isGeneric:
			return f'<{cls.__name__}>'
		return cls.__name__

	# @lru_cache(maxsize=512)
	def __findUnitClass__(cls, unit: str) -> Type[Measurement] | None:
		unitDict = cls.compatibleUnits
		# unitDict.update({v.__name__.lower(): v for v in unitDict.values() if v.__name__.lower() not in unitDict})
		closestMatch = get_close_matches(unit.lower(), (i.lower() for i in unitDict.keys()), 1, cutoff=0.7)
		if closestMatch:
			return unitDict[closestMatch[0]]
		return None

	@property
	def pluralName(self) -> Optional[str]:
		return getattr(self, '_pluralName', None)

	@property
	def pluralUnit(self) -> Optional[str]:
		return getattr(self, '_pluralUnit', None)

	@property
	def aliases(self) -> Set[str]:
		return getattr(self, '_aliases', set()) | {self.name, self.pluralName, self.pluralUnit} - {None, ''}

	def _limitFunc(cls, value: _T) -> float:
		if cls.isDerived:
			return cls.numeratorClass._limitFunc(value)
		return cls._limitFunc(value)


class SmartFloat(float, metaclass=MetaUnitClass):
	_limits = -inf, inf
	_precision: int = 3
	valuePrecision: int
	_max: int = 4
	_unit: Optional[str]
	_suffix: Optional[str]
	_decorator: Optional[str]
	_unitSpacer: Optional[Union[bool, str]]
	_title: Optional[str]
	_showUnit: Optional[bool]
	_shorten: Optional[bool]
	_kSeparator: Optional[str]
	_key: Optional[Union[str, Set[Union[str, Tuple[str]]]]]
	_sizeHint: Optional[str]
	_acceptedTypes: ClassVar[tuple] = (float, int, Decimal)
	__unitDict__: ChainMap[str, Type]

	@classmethod
	def noneToNan(cls, value):
		if not isinstance(value, float):
			try:
				value = float(value)
			except ValueError:
				value = nan
			except TypeError:
				value = nan
		return value

	def __new__(cls, value):
		return float.__new__(cls, value)

	def __init__(self, value):
		float.__init__(value)

	@staticmethod
	def intFloatLength(value: float) -> tuple[int, int]:
		value = float(value)
		d, *f = (len(i) for i in f'{abs(value):g}'.split('.'))
		return d, f[0] if f else 0

	def _string(
		self,
		shorten: bool = None,
		prefix: str = None,
		suffix: str = None,
		decorator: str = None,
		spacer: Union[bool, str] = None,
		unit: bool = None,
		maxLength: int = None,
		formatSpec: str = None,
	) -> str:
		if shorten is None:
			shorten = getattr(self, '_shorten', False)
		if prefix is None:
			prefix = getattr(self, '_prefix', '')
		if suffix is None:
			suffix = getattr(self, '_suffix', '')
		if decorator is None:
			decorator = getattr(self, '_decorator', '')
		if spacer is None:
			spacer = getattr(self, '_unitSpacer', '')
		elif spacer is True:
			spacer = getattr(self, '_spacer', ' ') or ' '
		else:
			spacer = ''
		if unit is None:
			unit = getattr(self, 'unit', True)
		if maxLength is None:
			maxLength = getattr(self, '_max', 4)
		if formatSpec is None:
			formatSpec = getattr(self, '_format', 'g')

		if shorten:
			c, valueFloat = 0, float(self)
			numberLength = len(str(int(valueFloat)))
			while numberLength > 3 and numberLength >= self._max:
				c += 1
				valueFloat /= 1000
				numberLength = len(str(int(valueFloat)))
			valueSuffix = ['', 'k', 'm', 'B'][c]
		else:
			valueFloat = float(self)
			c = False
			valueSuffix = ''
		if isnan(valueFloat):
			return '???'

		# TODO: Allow for precision to be overridden if number is scaled.  (10,110.0 becomes 10.11k instead of 10.1k)
		integerLength, floatingPointLength = self.intFloatLength(valueFloat)
		if floatingPointLength == 1 and valueFloat.is_integer():
			floatingPointLength = 0

		stringType = 'f'

		# Max amount of precision that can be displayed while keeping string under max length
		intAllowedPrecision = max(0, self._max - integerLength)
		precision = min(self._precision, integerLength)
		# Allow at least on level of precision if
		# Removed 1 if not decimal and c else decimal

		f'{prefix}{self:{formatSpec}}{valueSuffix}{suffix}{decorator}{spacer}{unit}'

	def __str__(self):
		return f'{self}'

	def __format__(self, formatSpec):
		convertTo = None
		compatible_units = type(self).compatibleUnits

		conversionSpecMatch = FormatSpec.conversion.search(formatSpec)
		unitSet = set(compatible_units) | {'auto'}
		if conversionSpecMatch:
			conversionSpecMatch = conversionSpecMatch.groupdict()
			parsedUnit = type(self).__parse_unit__(conversionSpecMatch['convertTo'])
			if isinstance(parsedUnit, str) and parsedUnit.lower() in unitSet:
				formatSpec = re.sub(fr'{conversionSpecMatch["fullmatch"]}', '', formatSpec)
				convertTo = ({parsedUnit} & unitSet).pop()
			elif isinstance(parsedUnit, Iterable) and hasattr(self, 'numerator') and any(i.lower() in unitSet for i in parsedUnit):
				formatSpec = re.sub(fr'{conversionSpecMatch["fullmatch"]}', '', formatSpec)
				convertTo = tuple(i.lower() for i in parsedUnit if i in unitSet)

		if precisionSpec := FormatSpec.precision.search(formatSpec):
			precisionSpec = precisionSpec.groupdict()
		else:
			precisionSpec = {'format_spec': ''}

		specParams = dict((i['key'], i['value']) for i in FormatSpec.params.finditer(formatSpec))
		for key, value in specParams.items():
			if number := FormatSpec.getNumber(value, strict=True):
				specParams[key] = number
			elif (boolValue := FormatSpec.getBool(value, strict=None)) is not None:
				specParams[key] = boolValue if boolValue is not None else empty

		paramsSpecConversionMatch = type(self).__parse_unit__(specParams.get('convert', None))
		paramsSpecConversion = next(iter({paramsSpecConversionMatch} & unitSet), None)
		if convertTo := (convertTo or paramsSpecConversion):
			if convertTo == 'auto':
				if (auto := getattr(self, 'auto', None)) is not None:
					value = auto
				else:
					raise NotImplementedError(f'No auto conversion for {self.name}')
			else:
				value = type(self).type.getClass(convertTo)(self)
		elif convertTo or paramsSpecConversionMatch:
			attemptedUnit = []
			if conversionSpecMatch:
				attemptedUnit.append(conversionSpecMatch['convertTo'])
			if paramsSpecConversionMatch:
				attemptedUnit.append(paramsSpecConversionMatch)
			compatibleUnits = sorted(i.unit for i in type(self).type.subTypes)
			raise NotImplementedError(f'{" or ".join(repr(u) for u in attemptedUnit)} is not a valid conversion for {self.name}.  Valid conversions are: {compatibleUnits!r}')
		else:
			value = self

		floatValue = fm if (fm := getattr(value, 'formatValue', None)) is not None else float(value)

		params = ChainMap(specParams)
		params.specParams = specParams
		params.default = value.defaultFormatParams
		params.maps.insert(1, params.default)

		shortened = False
		if params.get('shorten', False):
			if (auto := getattr(self, 'auto', None)) is not None:
				value = auto
				floatValue = float(value)
			else:
				c = f'{floatValue:,}'.count(',')
				if c:
					shortened = True
					floatValue /= 1000 ** c
					params['valueSuffix'] = ['', 'k', 'm', 'B'][c]

		# get the formatSpec from the params['format']
		formatString = params.get('format', self.__format_template__(params))
		params.formatString = formatString

		# get format specs from inside the param formatSpec
		if formatStringParams := FormatSpec.valueParams.search(formatString):
			formatStringParams = formatStringParams.groupdict()
			formatStringParams = {key: value for key, value in formatStringParams.items() if value is not None}
		else:
			formatStringParams = {}

		params.maps.insert(0, formatStringParams)
		params.formatStringParams = formatStringParams

		# get format specs from formatSpec

		precisionSpec = {key: value for key, value in precisionSpec.items() if value is not None}
		params.maps.insert(0, precisionSpec)
		params.precisionSpec = precisionSpec

		formatVars = {i['name']: i.groupdict() for i in FormatSpec.formatParams.finditer(formatString)}
		params.formatVars = formatVars

		params['value'] = floatValue
		params.maps.append(self.__dict__)
		params.dict = self.__dict__

		for key, opts in formatVars.items():
			spec = opts['spec']
			v = getFrom(key, value, self, params, locals(), globals())
			if v is UnsetKwarg:
				raise ValueError(f"Variable '{key}' is undefined")
			if spec and key not in {'format', 'self', 'value'}:
				v = f'{v:{spec}}'
			formatVars[key] = v

		params.maps.insert(0, formatVars)

		if params['type'] == 'g':
			p = int(params['precision'])
			if floatValue > 1:
				max_ = int(params['max'])
				intLength, valuePrecision = value.intFloatLength(floatValue)
				if p:
					totalLength = intLength + min(p, valuePrecision)
					params['precision'] = min(totalLength, max_) or 1
				else:
					if shortened:
						totalLength = min(intLength + valuePrecision, max_) - intLength
						params['precision'] = min(valuePrecision, totalLength) or 1
						params['type'] = 'f'
					else:
						params['precision'] = intLength or 1
			else:
				params['type'] = 'f'
				params['precision'] = p

		# params = value.__format_class__(formatSpec, params)
		formatString = params.formatString

		# params = {k: v if v is not None else '' for k, v in params.items() if not k.startswith('_')}
		# for k, v in params.items():
		# 	if k in {'precision', 'minwidth'} and v:
		# 		params[k] = int(v)

		params['value'] = self.__format_value__(params)

		formattedValue = formatString.format(**params)

		return formattedValue

	def __format_class__(self, formatSpec: str, formatParams: Mapping) -> dict:
		return formatParams

	def __format_value__(self, params: Mapping) -> str:
		return '{value:{fill}{align}{sign}{minwidth}.{precision}{type}}'.format(**params)

	def __format_template__(self, params: Mapping) -> str:
		formatTemplate = []
		if params.get('prefix', None):
			formatTemplate.append('{prefix}')
		value = params.get('value', None)
		if value is not False or value is not None:
			formatTemplate.append('{value}')
		if valueSuffix := params.get('valueSuffix', None):
			formatTemplate.append(f'{valueSuffix}')
		if params.get('decorator', None):
			formatTemplate.append('{decorator}')
		if params.get('showUnit', self.showUnit) and params.get('unit', self.unit) is not False:
			if params.get('unitSpacer', self.unitSpacer):
				formatTemplate.append('{unitSpacer}')
			formatTemplate.append('{unit}')
		if params.get('suffix', None):
			formatTemplate.append('{suffix}')

		return ''.join(formatTemplate)

	@property
	def defaultFormat(self) -> str:
		if self.showUnit:
			return "{value}{decorator}{unitSpacer}{unit}"
		return "{value}{decorator}"

	@property
	def defaultFormatParams(self):
		return {
			'leadingZero':  True,
			'trailingZero': True,
			'align':        '',
			'fill':         '',
			'sign':         '',
			'minwidth':     '',
			'precision':    self.precision,
			'value':        self,
			'type':         'g',
			'decorator':    self.decorator,
			'unitSpacer':   self.unitSpacer,
			'unit':         self.unit,
			'showUnit':     self.showUnit,
			'suffix':       self.suffix,
			'max':          self.max,
			'limits':       self._limits,
			'shorten':      self.shorten,
		}

	@property
	def properties(self) -> dict:
		attrs = {}
		if dimension := getattr(self, 'dimension', None):
			attrs['dimension'] = dimension
		if unit := getattr(self, 'unit', None):
			attrs['unit'] = unit
		if (numerator := getattr(self, 'numerator', None)) and (denominator := getattr(self, 'denominator', None)):
			attrs['numerator'] = numerator
			attrs['denominator'] = denominator
		if suffix := getattr(self, 'suffix', None):
			attrs['suffix'] = suffix
		if decorator := getattr(self, 'decorator', None):
			attrs['decorator'] = decorator
		if maxLength := getattr(self, 'max', None):
			attrs['maxLength'] = maxLength
		if precision := getattr(self, '_precision', None):
			attrs['precision'] = precision
		if showUnit := getattr(self, 'showUnit', None):
			attrs['showUnit'] = showUnit
		if shorten := getattr(self, 'shorten', None):
			attrs['shorten'] = shorten
		if (limits := getattr(self, '_limits', None)) and limits != (-inf, inf):
			attrs['limits'] = limits
		return attrs

	def __repr__(self):
		attrsString = ', '.join(f'{k}={str(v)}' for k, v in self.properties.items())
		return f'{type(self).__name__}(value={self.__repr_value__()}, {attrsString})'

	def __repr_value__(self) -> str:
		return f'{self: format:{"{value}{decorator}"}, type: g}'

	def __bool__(self):
		return super().__bool__() or bool(self.unit)

	def __float__(self) -> float:
		return super().__float__()

	def __hash__(self):
		return hash(round(self, max(self._precision, 1)))

	def __dir__(self):
		return list(self.properties) + ['__class__', '__module__', '__doc__', 'defaultFormat', 'defaultFormatParams']

	@property
	def valueUnset(self) -> bool:
		return self == nan

	@property
	def withUnit(self) -> str:
		return f'{self:showUnit: True}'

	@property
	def withoutUnit(self) -> str:
		return f'{self:showUnit: False}'

	@property
	def unit(self) -> str:
		return getattr(self, '_unit', '')

	@unit.setter
	def unit(self, value: str):
		self._unit = value

	@property
	def name(self):
		return getattr(self, '_name', type(self).__name__)

	@property
	def unitArray(self) -> list[str]:
		unit = self.unit
		return [unit] if unit is not None else []

	@property
	def suffix(self) -> Optional[str]:
		return getattr(self, '_suffix', '')

	@property
	def as_int(self) -> float:
		return self.__class__(round(self))

	@property
	def as_decorated_int(self) -> str:
		return f'{self:#.0g}'

	@property
	def name(self) -> str:
		return self.__class__.__name__

	@property
	def title(self) -> Optional[str]:
		return getattr(self, '_title', None)

	@title.setter
	def title(self, value):
		self._title = value

	@property
	def key(self) -> Optional[str]:
		return getattr(self, '_key', None)

	@key.setter
	def key(self, value: str):
		self._key = value

	@property
	def showUnit(self) -> bool:
		return getattr(self, '_showUnit', True)

	@showUnit.setter
	def showUnit(self, value):
		self._showUnit = value

	@property
	def decorator(self) -> str:
		return getattr(self, '_decorator', empty)

	@property
	def precision(self) -> int:
		_, valuePrecision = self.intFloatLength(self)
		return min(valuePrecision, getattr(self, '_precision', inf))

	@precision.setter
	def precision(self, value):
		self._precision = value

	@property
	def formatType(self) -> str:
		return 'g'

	@property
	def max(self) -> int:
		return self._max

	@max.setter
	def max(self, value: int):
		self._max = value

	@property
	def shorten(self) -> bool:
		return getattr(self, '_shorten', False)

	@shorten.setter
	def shorten(self, value):
		self._shorten = value

	@property
	def unitSpacer(self):
		spacer = getattr(self, '_unitSpacer', False)
		if isinstance(spacer, str):
			return spacer
		return " " if spacer else ""

	@unitSpacer.setter
	def unitSpacer(self, value):
		self._unitSpacer = value



from enum import Enum

from ..base.Decorators import UnitType
from . import Length, Time, DistanceOverTime

__all__ = ['Precipitation']


class Precipitation:
	class Type(Enum):
		NONE = 0
		Rain = 1
		Hail = 2
		Sleet = 3
		Snow = 4


@UnitType
class PrecipitationRate(DistanceOverTime, Precipitation, limit=(0.0, None)):
	Daily: type
	Hourly: type
	Minutely: type

	def __init__(self, numerator: Length, denominator: int = None, *args, **kwargs):
		if isinstance(denominator, int):
			denominator = Time.Hour(denominator)
		DistanceOverTime.__init__(self, numerator, denominator, *args, **kwargs)

	@property
	def fth(self):
		return Hourly(self._numerator.foot, self._denominator.hour)

	@property
	def inh(self):
		return Hourly(self._numerator.inch, self._denominator.hour)

	@property
	def ins(self):
		return PrecipitationRate(self._numerator.inch, self._denominator.second)

	@property
	def mmh(self):
		return Hourly(self._numerator.millimeter, self._denominator.hour)

	@property
	def mms(self):
		return PrecipitationRate(self._numerator.millimeter, self._denominator.second)

	@property
	def cmh(self):
		return Hourly(self._numerator.centimeter, self._denominator.hour)

	@property
	def mh(self):
		return Hourly(self._numerator.meter, self._denominator.hour)

	@property
	def daily(self):
		return Daily(self._numerator, self._denominator)

	@property
	def hourly(self):
		return Hourly(self._numerator, self._denominator)

	@property
	def minutely(self):
		return Minutely(self._numerator, self._denominator)

	@property
	def secondly(self):
		return Secondly(self.numerator, self.denominator)


class Daily(PrecipitationRate, denominator=Time.Day): ...


class Hourly(PrecipitationRate, denominator=Time.Hour): ...


class Minutely(PrecipitationRate, denominator=Time.Minute): ...


class Secondly(PrecipitationRate, denominator=Time.Second): ...


Precipitation.Rate = PrecipitationRate
Precipitation.Daily = Daily
Precipitation.Hourly = Hourly
Precipitation.Minutely = Minutely
Precipitation.Secondly = Secondly



from math import inf

from ..base import metric
from .temperature import Temperature

__all__ = ['Kelvin']


class Kelvin(Temperature, system=metric):
	_limits = (0, inf)
	_unit = 'k'

	def _kelvin(self):
		return self

	def _celsius(self, delta: bool = False):
		return float(self) - 273.15

	def _fahrenheit(self, delta: bool = False):
		if delta:
			return (float(self) - 273.15)*1.8
		else:
			return ((float(self) - 273.15)*1.8) + 32



from math import inf

from ..base import metric
from .temperature import Temperature

__all__ = ['Celsius']


class Celsius(Temperature, system=metric):
	_limits = (-273.15, inf)
	_unit = 'c'

	def _celsius(self, delta: bool = False):
		return self

	def _fahrenheit(self, delta: bool = False):
		if delta:
			return float(self)*1.8
		else:
			return (float(self)*1.8) + 32

	def _kelvin(self):
		return float(self) + 273.15



from datetime import timedelta as _td

__all__ = ['DistanceOverTime']

from ..base.Decorators import UnitType
from . import Length, Time, DerivedMeasurement


@UnitType
class DistanceOverTime(DerivedMeasurement, numerator=Length, denominator=Time):

	def __new__(cls, numerator: Length, denominator: Time = None):
		if isinstance(denominator, _td):
			denominator = Time.Second(denominator.total_seconds())
		return super().__new__(cls, numerator, denominator)

	def __init__(self, numerator: Length, denominator: Time = None):
		if isinstance(denominator, _td):
			denominator = Time.Second(denominator.total_seconds())
		super().__init__(numerator, denominator)

	@property
	def fts(self):
		return DistanceOverTime(self._numerator.ft, self._denominator.s)

	@property
	def mih(self):
		converted = MilesPerHour(self._numerator.mi, self._denominator.hr)
		return converted

	@property
	def inh(self):
		return DistanceOverTime(self._numerator.inch, self._denominator.hr)

	@property
	def ms(self):
		return DistanceOverTime(self._numerator.m, self._denominator.s)

	@property
	def mh(self):
		return DistanceOverTime(self._numerator.m, self._denominator.hr)

	@property
	def kmh(self):
		return DistanceOverTime(self._numerator.km, self._denominator.hr)

	@property
	def mmh(self):
		return DistanceOverTime(self._numerator.mm, self._denominator.hr)

	mph = mih
	fps = fts


class PerSecond(DistanceOverTime, numerator=Length, denominator=Time.Second):
	...


class PerMinute(DistanceOverTime, numerator=Length, denominator=Time.Minute):
	...


class PerHour(DistanceOverTime, numerator=Length, denominator=Time.Hour):
	...


class MilesPerHour(PerHour, numerator=Length.Mile, unit='mph'):
	...


class MetersPerSecond(PerSecond, numerator=Length.Meter):
	...


DistanceOverTime.PerSecond = PerSecond
DistanceOverTime.PerMinute = PerMinute
DistanceOverTime.PerHour = PerHour
DistanceOverTime.MilesPerHour = MilesPerHour
DistanceOverTime.MetersPerSecond = MetersPerSecond
DistanceOverTime.MetersPerHour = MetersPerSecond



from ..base.Decorators import Tiny, Small, Medium, Large
from ..base import Scale, ScalingMeasurement
from ..base import imperial
from .length import Length


class ImperialLength(ScalingMeasurement, Length, system=imperial, baseUnit='Foot'):
	class _Scale(Scale):
		Line = 1
		Inch = 12
		Foot = 12
		Yard = 3
		Mile = 1760
		Base = 'Foot'

	def _meter(self):
		return self.changeScale(self._Scale.Foot) * 0.3048


@Tiny
class Line(ImperialLength):
	_unit = 'ln'


@Small
class Inch(ImperialLength):
	_unit = 'in'


@Medium
class Foot(ImperialLength, pluralName='Feet'):
	_unit = 'ft'


@Medium
class Yard(ImperialLength):
	_unit = 'yd'


@Large
class Mile(ImperialLength):
	_unit = 'mi'
	_precision = 1



from ..base import Dimension

__all__ = ['Mass']


class Mass(metaclass=Dimension, symbol='M'):
	Gram: type
	Microgram: type
	Milligram: type
	Kilogram: type
	Dram: type
	Ounce: type
	Pound: type
	Hundredweight: type
	Ton: type

	@property
	def microgram(self):
		from . import Microgram
		return Microgram(self)

	@property
	def milligram(self):
		from . import Milligram
		return Milligram(self)

	@property
	def gram(self):
		from . import Gram
		return Gram(self)

	@property
	def kilogram(self):
		from . import Kilogram
		return Kilogram(self)

	@property
	def dram(self):
		from . import Dram
		return Dram(self)

	@property
	def ounce(self):
		from . import Ounce
		return Ounce(self)

	@property
	def pound(self):
		from . import Pound
		return Pound(self)

	ug = microgram
	mg = milligram
	g = gram
	kg = kilogram
	oz = ounce
	lbs = pound



from datetime import timedelta, datetime

from math import inf

__all__ = ['Time']

from ..base.Decorators import UnitType
from ..base import Dimension, both
from ..base import ScalingMeasurement, Scale


@UnitType
class Time(ScalingMeasurement, metaclass=Dimension, system=both, symbol='T', baseUnit='Second'):
	_limits = -inf, inf
	Millisecond: type
	Second: type
	Minute: type
	Hour: type
	Day: type
	Week: type
	Month: type
	Year: type
	Decade: type
	Century: type
	Millennia: type

	_acceptedTypes = (timedelta,)

	class _Scale(Scale):
		Millisecond = 1/1000
		Second = 1
		Minute = 60
		Hour = 60
		Day = 24
		Year = 365
		Decade = 10
		Century = 10
		Millennia = 10
		Base = 'Second'
		Week = 7.0*float(Day*Hour*Minute*Second)
		Month = 30.436875*float(Day*Hour*Minute*Second)

	def _convert(self, other):
		if isinstance(other, datetime):
			other = Second(other.timestamp() - datetime.now().timestamp())
		if isinstance(other, timedelta):
			other = Second(other.total_seconds())
		return super()._convert(other)

	def _millisecond(self):
		return self.changeScale(self.scale.Millisecond)

	def _second(self):
		return self.changeScale(self.scale.Second)

	def _minute(self):
		return self.changeScale(self.scale.Minute)

	def _hour(self):
		return self.changeScale(self.scale.Hour)

	def _day(self):
		return self.changeScale(self.scale.Day)

	def _year(self):
		return self.changeScale(self.scale.Year)

	def _month(self):
		return float(Month(self))

	def _decade(self):
		return self.changeScale(self.scale.Decade)

	def _century(self):
		return self.changeScale(self.scale.Century)

	def _millennia(self):
		return self.changeScale(self.scale.Millennia)

	@property
	def auto(self):
		if self._second() < 60 and self.scale <= self._Scale.Second:
			return self.s
		elif self._minute() < 60 and self.scale <= self._Scale.Minute:
			return self.min
		elif self._hour() < 24 and self.scale <= self._Scale.Hour:
			return self.hour
		elif self._day() < 7 and self.scale <= self._Scale.Year:
			return self.week
		elif self._day() < 30 and self.scale <= self._Scale.Year:
			return self.month
		return self

	@property
	def autoAny(self):
		if abs(self._second()) < 60:
			return self.s
		elif abs(self._minute()) < 60:
			return self.min
		elif abs(self._hour()) < 24:
			return self.hour
		elif abs(self._day()) < 2:
			return self.day
		elif abs(self._day()) < 7:
			return self.week
		elif abs(self._day()) < 30:
			return self.month
		elif abs(self._year()) > 1:
			return self.year
		return self

	@property
	def millisecond(self):
		return Millisecond(self)

	@property
	def second(self):
		return Second(self)

	@property
	def minute(self):
		return Minute(self)

	@property
	def hour(self):
		return Hour(self)

	@property
	def day(self):
		return Day(self)

	@property
	def week(self):
		return Week(self)

	@property
	def month(self):
		return Month(self)

	@property
	def year(self):
		return Year(self)

	@property
	def decade(self):
		return Decade(self)

	@property
	def century(self):
		return Century(self)

	@property
	def millennia(self):
		return Millennia(self)

	## abbreviations
	y = year
	d = day
	min = minute
	m = minute
	hr = hour
	h = hour
	s = second
	sec = second
	ms = millisecond


class Millisecond(Time):
	_unit = 'ms'


class Second(Time, alias='sec'):
	_unit = 's'


class Minute(Time):
	_precision = 1
	_unit = 'min'


class Hour(Time):
	_unit = 'hr'


class Day(Time):
	_unit = 'd'


class Week(Time):
	_unit = 'wk'


class Month(Time):
	_unit = 'mth'


class Year(Time):
	_unit = 'yr'


class Decade(Time):
	_unit = 'dec'


class Century(Time):
	_unit = 'cen'


class Millennia(Time):
	_unit = 'mel'


Time.Millisecond = Millisecond
Time.Second = Second
Time.Minute = Minute
Time.Hour = Hour
Time.Day = Day
Time.Week = Week
Time.Month = Month
Time.Year = Year
Time.Decade = Decade
Time.Century = Century
Time.Millennia = Millennia



from ..base.Decorators import UnitType
from ..base import Scale
from . import ScalingMeasurement

__all__ = ['PartsPer']


@UnitType
class PartsPer(ScalingMeasurement, baseClass='Hundred'):
	_precision = 0

	class _Scale(Scale):
		Hundred = 100
		Thousand = 10
		TenThousand = 10
		HundredThousand = 10
		Million = 10
		Billion = 1000
		Trillion = 1000
		Base = 'Hundred'
		Precent = 100.

	@property
	def hundred(self):
		return Hundred(self)
	precent=hundred

	@property
	def thousand(self):
		return Thousand(self)

	@property
	def tenThousand(self):
		return TenThousand(self)

	@property
	def hundredThousand(self):
		return HundredThousand(self)

	@property
	def million(self):
		return Million(self)
	ppm = million

	@property
	def billion(self):
		return Billion(self)
	ppb = billion

	@property
	def trillion(self):
		return Trillion(self)
	ppt = trillion


class Hundred(PartsPer):
	_precision = 2
	_max = 3
	_unit = '%'


Percent = Hundred


class Thousand(PartsPer):
	_unit = '‰'


class TenThousand(PartsPer):
	_unit = '‱'


class HundredThousand(PartsPer):
	_unit = 'pcm'


class Million(PartsPer):
	_unit = 'ppm'


class Billion(PartsPer):
	_unit = 'ppb'


class Trillion(PartsPer):
	_unit = 'ppt'


PartsPer.Hunderd = Hundred
PartsPer.Precent = Hundred
PartsPer.Thousand = Thousand
PartsPer.TenThousand = TenThousand
PartsPer.HundredThousand = HundredThousand
PartsPer.Million = Million
PartsPer.Billion = Billion
PartsPer.Trillion = Trillion



from abc import abstractmethod
from enum import Enum, EnumMeta
from typing import Optional, Type

from .. import errors
from . import Measurement, DerivedMeasurement

__all__ = ['ScalingMeasurement', 'SystemVariant', 'Scale']


class AddressableEnum(EnumMeta):

	def __getitem__(self, indexOrName):
		if isinstance(indexOrName, str):
			return super().__getitem__(indexOrName)
		elif isinstance(indexOrName, int) and indexOrName < super().__len__():
			return self._list[indexOrName]
		elif isinstance(indexOrName, slice):
			indices = range(*indexOrName.indices(len(self._list)))
			return [self._list[i] for i in indices]

	@property
	def _list(self):
		return list(self)


class Indexer:
	# Todo: this could probably be moved inside or declared as a class variable of Scale

	i = 0
	lastClass: object = None

	@classmethod
	def get(cls, caller):
		cls.i, cls.lastClass = (0, caller) if not caller == cls.lastClass else (cls.i + 1, caller)
		return cls.i


class Scale(Enum, metaclass=AddressableEnum):

	def __new__(cls, *args):
		if isinstance(args[0], int):
			obj = object.__new__(cls)
			obj._value_ = Indexer.get(cls)
			obj._mul_ = args[0]
			obj.isVariant = False
		elif isinstance(args[0], float):
			obj = object.__new__(cls)
			obj._value_ = Indexer.get(cls)
			obj._mul_ = args[0]
			obj.isVariant = True
		elif isinstance(args[0], str):
			obj = object.__new__(cls)
			obj._value_ = getattr(cls, args[0]).index
			obj._mul_ = getattr(cls, args[0]).value
			obj.isVariant = False
		else:
			obj = None

		return obj

	def __str__(self):
		return str(self.value)

	# this makes sure that the description is read-only
	@property
	def index(self):
		if self.isVariant:
			return self.Base.index + 1
		return self._value_

	@property
	def value(self):
		return self._mul_

	def __repr__(self):
		return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, self._mul_)

	def __mul__(self, other):
		if isinstance(other, self.__class__):
			variant = self.isVariant or other.isVariant
			val = 1
			array = self.__class__[min(self.index, other.index) + 1: max(self.index, other.index) + 1]
			array.sort(key=lambda x: x.index)
			array = array[:-1] if variant else array
			for i in array:
				val *= i.value
			if variant:
				val *= self.value if self.isVariant else other.value
			return val

	def __gt__(self, other):
		if isinstance(other, self.__class__):
			return self.index > other.index
		else:
			return self.index > int(other)

	def __lt__(self, other):
		if isinstance(other, self.__class__):
			return self.index < other.index
		else:
			return self.index < int(other)

	def __eq__(self, other):
		if isinstance(other, self.__class__):
			return self.index == other.index
		else:
			return self.index == int(other)

	def __le__(self, other):
		if isinstance(other, self.__class__):
			return self.index <= other.index
		else:
			return self.index <= int(other)

	def __ge__(self, other):
		if isinstance(other, self.__class__):
			return self.index >= other.index
		else:
			return self.index >= int(other)


class BaseUnit:

	@abstractmethod
	def changeSystem(self, system):
		pass


class ScalingMeasurement(Measurement):
	_baseUnit: Type[Measurement]
	_Scale: Scale = None

	def __init_subclass__(cls, **kwargs):
		baseUnit = kwargs.get('baseUnit', None)
		if baseUnit:
			cls._baseUnitRef = baseUnit
			cls._system = cls
			cls._systemName = kwargs.get('system', 'mixed')
		else:
			baseUnitRef = getattr(cls, '_baseUnitRef', None)
			if baseUnitRef == cls.__name__:
				cls._system._baseUnit = cls
		super().__init_subclass__(**kwargs)

	def __new__(cls, value, *args, **kwargs):
		if isinstance(value, DerivedMeasurement):
			return value
		valueCls = type(value)

		if not issubclass(valueCls, Measurement) and issubclass(valueCls, (float, int)):
			value = float(value)
			return super().__new__(cls, value, *args, **kwargs)

		if isinstance(value, ScalingMeasurement) and (not cls._baseUnit or not value.__class__._baseUnit):
			'''For this to work each unit class must have a _baseUnit defined for each scale'''
			raise errors.Unit.NoBaseUnitDefined(cls)

		sameDimension = cls.dimension is valueCls.dimension
		sameSystem = cls.system is valueCls.system
		sameUnit = cls.unit is valueCls.unit

		if sameUnit and issubclass(valueCls, cls.type):
			return Measurement.__new__(cls, float(value), *args, **kwargs)

		if sameSystem and sameDimension:
			if isinstance(value, ScalingMeasurement) and not isinstance(value, value._baseUnit):
				value = cls.changeScale(value, cls._Scale.Base)
			if cls is cls._baseUnit:
				return cls.__new__(cls, value)
			else:
				value = cls._baseUnit.changeScale(value, cls._Scale[cls.__name__], cls._Scale.Base)
			return cls.__new__(cls, value)

		# If values are cousins initiate with values base sibling causing a recursive call to __new__
		elif sameDimension:
			if not isinstance(value, value._baseUnit):
				value = value.toBaseUnit()
			if (converter := getattr(value, f'_{cls._baseUnitRef.lower()}', None)) is not None:
				value = cls._baseUnit(converter())
			return cls.__new__(cls, value)

		raise errors.Conversion.BadConversion(cls.__name__, value.__class__.__name__)

	def changeScale(self, newUnit: Scale, scale: Scale = None) -> Optional[float]:
		scale = getattr(self, 'scale', scale)
		multiplier = scale*newUnit
		if scale > newUnit:
			return float(self)*multiplier
		else:
			return float(self)/multiplier

	def toBaseUnit(self) -> Measurement:
		return self._baseUnit(self.changeScale(self._Scale.Base))

	@property
	def scale(self):
		return self._Scale[self.name]

	@property
	def isSystemVariant(self):
		return issubclass(self.__class__, SystemVariant)


class SystemVariant:
	_multiplier: float = 1.0



from ..base.Decorators import Tiny, Small, Large, Huge
from ..base import Scale, ScalingMeasurement, Dimension, imperial
from .mass import Mass

__all__ = ['Dram', 'Ounce', 'Pound', 'Hundredweight', 'Ton']


class Imperial(ScalingMeasurement, Mass, metaclass=Dimension, system=imperial, baseUnit='Pound'):
	class _Scale(Scale):
		Dram = 1
		Ounce = 16
		Pound = 16
		Hundredweight = 100
		Ton = 20
		Base = 'Pound'

	def _gram(self):
		return self.changeScale(self._Scale.Ounce) * 28.349523125


@Tiny
class Dram(Imperial):
	_unit = 'dr'


@Small
class Ounce(Imperial):
	_unit = 'oz'


@Small
class Pound(Imperial):
	_unit = 'lbs'


@Large
class Hundredweight(Imperial):
	_unit = 'cwt'


@Huge
class Ton(Imperial):
	_id = 'tUS'
	_unit = 't'



from ..base.Decorators import UnitType
from . import Length, Time, DistanceOverTime, Direction

__all__ = ['Wind']


@UnitType
class Wind(DistanceOverTime):
	# TODO: Add support for setting both speed and direction
	__direction: Direction = None

	@property
	def direction(self):
		return self.__direction

	@direction.setter
	def direction(self, value):
		if not isinstance(value, Direction) and isinstance(value, (float, int)):
			value = Direction(value)
		self.__direction = value


class PerSecond(Wind, DistanceOverTime.PerSecond):
	...


class PerMinute(Wind, DistanceOverTime.PerMinute):
	...


class PerHour(Wind, DistanceOverTime.PerHour):
	...


class MilesPerHour(Wind, DistanceOverTime.MilesPerHour):
	...


class MetersPerSecond(Wind, DistanceOverTime.MetersPerSecond):
	...



import logging
import re
from difflib import get_close_matches, SequenceMatcher
from functools import lru_cache
from inspect import getsource, getfile, getmodule
from itertools import product
from types import FunctionType
from typing import Hashable, Type, Union, Final, Any, Mapping, Callable, Set, Tuple, Optional, NamedTuple

log = logging.getLogger(__name__)

numeric = re.compile(r'^[-+]?[0-9]*\.?[0-9]+$')


@lru_cache(maxsize=128)
def loadUnitLocalization(measurement: Type['Measurement'], config):
	if not isinstance(measurement, type):
		measurement = type(measurement)
	unitName = measurement.type.name.lower()
	unitType = measurement.Generic.name.lower()
	match = get_close_matches(unitName, config.localUnits.keys(), n=1, cutoff=0.85) or get_close_matches(unitType, config.localUnits.keys(), n=1, cutoff=0.8)
	return config.localUnits[match[0]] if match else None


# possibleTypes = (unitName.lower(), toCamelCase(unitName), unitName)
# for classType in {*possibleNames, *possibleTypes}:
# 	if classType in config.localUnits:
# 		return config.localUnits[classType]
# else:
# 	log.warning(f'{measurement._type.__name__} has no type defined')


def toCamelCase(string, titleCase: bool = False) -> str:
	string = (string.lower() if string.isupper() else string)
	for char in ['-', ' ', '.', '_']:
		string.replace(char, '')
	return string[0].upper() if titleCase else string[0].lower() + string[1:]


def convertString(value: str) -> Union[int, float, str, bool]:
	if value is None:
		return value
	if value.isnumeric():
		value = float(value)
		if value.is_integer():
			value = int(value)
		return value
	if value == 'True':
		value = True
	elif value == 'False':
		value = False
	return value


ScoredMatch = NamedTuple('Match', [('ratio', 'float'), ('value', 'str')])

scored_get_close_matches: FunctionType
sourceFile = getfile(get_close_matches)
get_close_matches_code = getsource(get_close_matches)
get_close_matches_code = get_close_matches_code.replace('def get_close_matches', 'def scored_get_close_matches')
get_close_matches_code = get_close_matches_code.replace('return [x for score, x in result]', 'return tuple(ScoredMatch(k, v) for (k, v) in result)')
exec(compile(get_close_matches_code, sourceFile, 'exec'), {**getmodule(get_close_matches, sourceFile).__dict__, **locals()}, globals())


def best_match(word, possibilities, cutoff=0) -> Optional[str]:
	"""
	Returns the best guess at a match to the word.
	"""
	guesses = get_close_matches(word, possibilities, n=1, cutoff=cutoff)
	return guesses[0] if guesses else None


class CaseInsensitiveKey(str):

	def __hash__(self):
		return hash(self.lower())

	def __eq__(self, other):
		return self.lower() == other.lower()


def parseString(item: str):
	key, value = item.split('=')
	value = convertString(value)
	return f'_{key}', value


def strToDict(string: str, cls: type) -> type:
	conf = [parseString(a) for a in [(y.strip(' ')) for y in string.split(',')]]
	for item in conf:
		setattr(cls, *item)
	return cls


def setPropertiesFromConfig(cls, config) -> None:
	"""
	Modifies the class in place to add the properties defined in the config.
	:param cls: Class to modify
	:type cls: Type[Measurement]
	:param config: Config to use
	:type config: Config
	"""
	possibleNames = [cls.__name__.lower(), cls._unit, toCamelCase(cls.__name__), cls.__name__]
	if hasattr(cls, '_type') and cls._type is not None and not isinstance(cls._type, str):
		possibleTypes = [cls._type.__name__.lower(), toCamelCase(cls._type.__name__), cls._type.__name__]
		for classType in possibleTypes:
			try:
				cls = strToDict(config.unitProperties[classType], cls)
				config.configuredUnits[cls.__name__] = cls
				break
			except KeyError:
				pass
		else:
			log.debug(f'{cls.__name__} has no type defined')

	for name in possibleNames:
		try:
			cls = strToDict(config.unitProperties[name], cls)
			config.configuredUnits[name] = cls
			break
		except KeyError:
			pass


def getPropertiesFromConfig(name: str, cls: dict, config) -> dict:
	"""
	Modifies the class in place to add the properties defined in the config.
	:param config: Config to use
	:type config: Config
	"""
	possibleNames = [name.__name__.lower(), cls.get('_unit'), toCamelCase(cls.__name__), name]
	# if hasattr(cls, '_type') and cls._type is not None and not isinstance(cls._type, str):
	# 	possibleTypes = [cls._type.__name__.lower(), toCamelCase(cls._type.__name__), cls._type.__name__]
	# 	for classType in possibleTypes:
	# 		try:
	# 			cls = strToDict(config.unitProperties[classType], cls)
	# 			config.configuredUnits[cls.__name__] = cls
	# 			break
	# 		except KeyError:
	# 			pass
	# 	else:
	# 		log.debug(f'{cls.__name__} has no type defined')
	result = {}
	for name in possibleNames:
		try:
			cls = strToDict(config.unitProperties[name], cls)
			config.configuredUnits[name] = cls
			break
		except KeyError:
			pass


class HashSlice:
	__slots__ = ('start', 'stop', 'step')
	start: Union[Hashable, type]
	stop: Union[Hashable, type]
	step: Union[Hashable, type]

	def __init__(self, start: Union[Hashable, slice], stop: Hashable = None, step: Hashable = None):
		if isinstance(start, (slice, HashSlice)):
			stop = start.stop
			step = start.step
			start = start.start
		elif isinstance(start, tuple) and stop is None and step is None:
			tupleLength = len(start)
			if tupleLength == 1:
				start = start[0]
			elif tupleLength == 2:
				start, stop = start
			elif tupleLength == 3:
				start, stop, step = start
			else:
				raise ValueError(f'Invalid start tuple length: {tupleLength}')
		self.start = start
		self.stop = stop
		self.step = step

	def __iter__(self) -> iter:
		return iter((self.start, self.stop, self.step))

	def __repr__(self) -> str:
		return f"HashSlice[{self.start or ''}:{self.stop or ''}:{self.step or ''}]"

	def __str__(self) -> str:
		return f"[{self.start or ''}:{self.stop or ''}:{self.step or ''}]"

	def __eq__(self, other: object) -> bool:
		if hasattr(other, 'start') and hasattr(other, 'stop') and hasattr(other, 'step'):
			return self.start == other.start and self.stop == other.stop and self.step == other.step
		if self.stop is None and self.step is None:
			return self.start == other
		return False

	def __hash__(self) -> int:
		values = [self.start, self.stop, self.step]
		if self.step is None:
			values.pop()
		if self.stop is None and self.step is None:
			values.pop()
		if len(values) == 1:
			return hash(values[0])
		return hash(tuple(values))

	@property
	def isSlice(self) -> bool:
		return self.stop is not None


def modifyCase(string: str, titleCase: bool = True, joiner: str = '') -> str:
	if string.isupper():
		return string
	stringParts = re.findall(r'((?=[A-Z\_]?)[A-Z]?[A-Za-z][a-z]+)', string) or []
	if titleCase and stringParts:
		return joiner.join(i.title() for i in stringParts)
	return joiner.join(stringParts)


@lru_cache(maxsize=1024)
def pluralize(word: str) -> str:
	if re.search('[sxz]$', word):
		if len(word) <= 2:
			return word
		return re.sub('$', 'es', word)
	elif re.search('[^aeioudgkprt]h$', word):
		return re.sub('$', 'es', word)
	elif re.search('[aeiou]y$', word):
		return re.sub('y$', 'ies', word)
	else:
		return f'{word}s'


class Infix:
	def __init__(self, function=lambda x, y: x if x is not None else y):
		self.function = function

	def __ror__(self, other):
		return Infix(lambda x, self=self, other=other: self.function(other, x))

	def __or__(self, other):
		return self.function(other)

	def __rlshift__(self, other):
		return Infix(lambda x, self=self, other=other: self.function(other, x))

	def __rshift__(self, other):
		return self.function(other)

	def __call__(self, value1, value2):
		return self.function(value1, value2)

	def __rmatmul__(self, other):
		return Infix(lambda x, self=self, other=other: self.function(other, x))

	def __matmul__(self, other):
		return self.function(other)


class _Empty:

	def __bool__(self) -> bool:
		return True

	def __str__(self) -> str:
		return ''

	def __repr__(self) -> str:
		return ''

	def __iter__(self) -> iter:
		return iter(())

	def __getitem__(self, item: object) -> object:
		return self

	def __getattr__(self, item):
		return self

	def __call__(self, *args, **kwargs):
		return self

	def __len__(self) -> int:
		return 0

	def __contains__(self, item: object) -> bool:
		return False

	def __int__(self) -> int:
		return 0

	def __float__(self) -> float:
		return 0.0


NotNone: Final = Infix()
empty: Final = _Empty()
UnsetKwarg: Final = object()
Unset: Final = object()


def getFrom(
	key: Hashable | str | Tuple[Hashable | str, ...],
	*objs: Mapping | object,
	default: Any = UnsetKwarg,
	expectedType: type | Tuple[type, ...] = object,
	factory: Callable[[Any], Any] = lambda x: x,
	ignore: Set[Any] = None,
	findAll: bool = False,
	pop: bool = False,
):
	"""
	Returns the first acceptable attr or value from a list of objects
	:param objs: The mapping or object to search.
	:type objs: Mapping | object
	:param key: The keys to search for.
	:type key: Hashable | str
	:param default: The default value to return if the key is not found.
	:type default: Any
	:return: The value of the key in the mapping or the default value.
	:rtype: Any
	"""
	if ignore is None:
		ignore = set()
	if not isinstance(key, tuple):
		key = (key,)
	if findAll:
		found = []
	for obj, subKey in product(objs, key):
		if isinstance(obj, Mapping):
			value = obj.get(subKey, default)
		else:
			value = getattr(obj, subKey, default)
		if value is not UnsetKwarg and value not in ignore and isinstance(value, expectedType):
			if pop and (_pop := getattr(obj, 'pop', None)) is not None:
				_pop(subKey)
			if not findAll:
				return factory(value)
			found.append(factory(value))
	if findAll:
		return found
	elif default is UnsetKwarg:
		raise KeyError(key)
	return default



import re
from collections import ChainMap

from functools import lru_cache, cached_property
import logging
from datetime import datetime
from typing import Callable, ClassVar, Dict, List, Optional, Type, Union, Final, Literal, Iterable

__all__ = ['Measurement', 'DerivedMeasurement', 'Dimension', 'metric', 'imperial', 'both', 'Dimensionless', 'Quantity', 'Index', 'NonPlural']

from .. import errors
from . import SmartFloat, FormatSpec, MetaUnitClass
from ..utils import HashSlice
from ..config import config

log = logging.getLogger('WeatherUnits')


class Measurement(SmartFloat):
	__unitDict__ = ChainMap()
	_derived: bool
	_type: type
	_updateFunction: Optional[Callable]
	_timestamp: datetime
	_indoor: bool
	_calculated: bool
	_category: str
	_subTypes: ClassVar[Dict[str, Type['Measurement']]]

	def __init_subclass__(cls, **kwargs):
		if kwargs.get('type', False):
			cls._type = cls

	def __new__(cls, value, title: str = None, key: str = None, timestamp: datetime = None, **kwargs):
		value = sorted((*cls.limits, float(value)))[1]
		return SmartFloat.__new__(cls, value)

	def __class_getitem__(cls, item):
		if (underItem := f'_{item}') in cls.__annotations__:
			return cls.__annotations__[underItem]
		if item in cls.__dict__:
			return cls.__dict__[item]
		if isinstance(item, str):
			item = cls.__findUnitClass__(item)
		if isinstance(item, type) and issubclass(item, Measurement):
			if item in cls.subTypes:
				return item
			parentType = cls.type
			name = f'{parentType.__name__}[{item.__name__}]'
			newCls = type(name, (item, cls,), {})
			return newCls

		raise TypeError(f'{cls.__name__} does not support {item}')

	@classmethod
	def getClass(cls, unit: str) -> Type['Measurement'] | None:
		if isinstance(unit, Iterable) and not isinstance(unit, str):
			unit = ''.join(unit)
		return cls.__findUnitClass__(unit)

	def __init__(self, value, title: str = None, key: str = None, timestamp: datetime = None, category: str = None):
		if isinstance(value, Measurement):
			if value._timestamp and timestamp is None:
				timestamp = value._timestamp
		if title:
			self._title = title
		if key:
			self._key = key
		self.category = category

		self._timestamp = timestamp

		SmartFloat.__init__(self, value)

	@property
	def category(self):
		if (category := getattr(self, '_category', None)) is None:
			return self.__class__.__name__.lower()
		return self._category

	@category.setter
	def category(self, value):
		self._category = value

	@property
	def localize(self):
		if self.convertible:
			try:
				if (newUnit := type(self).localizedUnit) is not None:
					return newUnit(self)
				return self
			except AttributeError or KeyError as e:
				errors.Conversion.BadConversion("Unable to get localized type for {}".format(self.name), e)
		else:
			return self

	@property
	def convertible(self) -> bool:
		return getattr(self, '_convertable', True)

	@property
	def str(self):
		return str(self)

	@property
	def type(self) -> type:
		return type(self).type

	@property
	def timestamp(self) -> datetime:
		return self._timestamp

	@property
	def updateFunction(self):
		return self._updateFunction

	@updateFunction.setter
	def updateFunction(self, function: callable):
		self._updateFunction = function

	@property
	def indoor(self) -> bool:
		return self._indoor

	@indoor.setter
	def indoor(self, value: bool):
		self._indoor = value

	@property
	def calculated(self):
		return self._calculated

	@calculated.setter
	def calculated(self, value):
		self._calculated = value

	def _convert(self, other):
		if isinstance(other, self.type):
			return self.__class__(other)
		if isinstance(other, (float, int)):
			return self.__class__(other)
		else:
			return other

	@cached_property
	def valuePrecision(self) -> int:
		_, value = self.intFloatLength(float(self))
		return value

	def transform(self, other, transformTo: type = None):
		if isinstance(other, type) and issubclass(other, Measurement):
			other = other(self)
		nonTransferred = ['_unit', '_suffix', '_scale', '_denominator', '_numerator']
		if self.type != other.type:
			log.warning(f'{self.withUnit} and {other.withUnit} are not identical types, this may cause issues')
		other.__dict__.update({key: value for key, value in self.__dict__.items() if key not in nonTransferred and value is not None})
		if getattr(other, '_updateFunction', None):
			other._updateFunction(other)
		return other

	def __wrapOther(self, other) -> float:
		if type(other) == type(self):
			pass
		elif isinstance(other, self.type):
			if self.unit != other.unit:
				other = self._convert(other)
		elif isinstance(other, (float, int)):
			return self.__class__(other)
		elif isinstance(other, self._acceptedTypes):
			other = self._convert(other)
		elif isinstance(other, str):
			if (other := FormatSpec.number.search(other)) is not None:
				other = other.groupdict()['number']
		return float(other)

	def __eq__(self, other):
		otherPrecision = getattr(other, 'valuePrecision', self.valuePrecision)
		other = self.__wrapOther(other)
		precision = min(self.valuePrecision, otherPrecision)
		selfVal = round(float(self), precision)
		other = round(float(other), precision)
		return selfVal == other

	def __getitem__(self, item):
		try:
			return self.__getattribute__(item)
		except AttributeError:
			if isinstance(item, str):
				return type(self)[item](self)
			raise errors.Conversion.UnknownUnit(self, item)

	def __mul__(self, other):
		other = self.__wrapOther(other)
		return type(self)(super().__mul__(other))

	def __add__(self, other):
		other = self.__wrapOther(other)
		return type(self)(super().__add__(other))

	def __radd__(self, other):
		other = self.__wrapOther(other)
		return type(self)(super().__radd__(other))

	def __sub__(self, other):
		other = self.__wrapOther(other)
		return type(self)(super().__sub__(other))

	def __rsub__(self, other):
		other = self.__wrapOther(other)
		return type(self)(super().__rsub__(other))

	def __pow__(self, power, modulo=None):
		other = self.__wrapOther(power)
		return type(self)(super().__pow__(other, modulo))

	def __truediv__(self, other):
		if isinstance(self, DerivedMeasurement) and not isinstance(other, DerivedMeasurement):
			numerator = type(self.numerator)(other)
			value = numerator/other
			return type(self)(value, self.denominator.__class__(1))

		other = self.__wrapOther(other)
		return type(self)(super().__truediv__(other))

	def __lt__(self, other):
		other = self.__wrapOther(other)
		return super().__lt__(other)

	def __gt__(self, other):
		other = self.__wrapOther(other)
		return super().__gt__(other)

	def __ge__(self, other):
		otherPrecision = getattr(other, 'valuePrecision', self.valuePrecision)
		other = self.__wrapOther(other)
		precision = min(self.valuePrecision, otherPrecision)
		selfVal = round(float(self), precision)
		other = round(float(other), precision)
		return selfVal >= other

	def __le__(self, other):
		otherPrecision = getattr(other, 'valuePrecision', self.valuePrecision)
		other = self.__wrapOther(other)
		precision = min(self.valuePrecision, otherPrecision)
		selfVal = round(float(self), precision)
		other = round(float(other), precision)
		return selfVal <= other

	def __abs__(self):
		return type(self)(super().__abs__())

	def __hash__(self):
		return hash(round(self, max(self._precision, 1)))


class DimensionlessMeta(MetaUnitClass):

	def __new__(cls, name, bases, attrs, **kwargs):
		return super().__new__(cls, name, bases, attrs, **kwargs)

	@property
	def type(cls):
		return getattr(cls, '_type', None) or next((base for base in cls.__mro__ if Dimensionless in base.__bases__), cls)

	@property
	def isGeneric(cls):
		return cls is Dimensionless

	@property
	def dimension(self):
		return Dimensionless


class Dimensionless(SmartFloat.__Dimensionless__, Measurement, metaclass=DimensionlessMeta):
	pass


class NonPlural(SmartFloat.__NonPlural__, Measurement):
	pass


class Quantity(Dimensionless):
	_max = 3
	_precision = 0


class Index(Dimensionless, NonPlural):
	_max = 3
	_precision = 0


class DerivedMeasurementMeta(MetaUnitClass):
	numerator: Type[Measurement]
	_numerator: Type[Measurement]

	denominator: Type[Measurement]
	_denominator: Type[Measurement]

	def __new__(mcs, *args, numerator: Type[Measurement] = None, denominator: Type[Measurement] = None, **kwargs):
		name, bases, attrs, *args = args
		if numerator is None:
			numerator = next((j for i in bases if (j := getattr(i, 'numerator', None)) is not None), None)
		else:
			attrs['_numerator'] = numerator
			annotations = attrs.get('__annotations__', {})
			annotations['numerator'] = Type[numerator]
			attrs['__annotations__'] = annotations

		if denominator is None:
			denominator = next((base_d for base in bases if (base_d := getattr(base, 'denominator', None)) is not Measurement), None)
		else:
			attrs['_denominator'] = denominator
			annotations = attrs.get('__annotations__', {})
			annotations['denominator'] = Type[denominator]
			attrs['__annotations__'] = annotations

		return super().__new__(mcs, name, bases, attrs, **kwargs)

	@property
	def isGeneric(cls):
		return cls.numerator.isGeneric or cls.denominator.isGeneric

	@property
	def numerator(self) -> Type[Measurement]:
		return getattr(self, '_numerator', None)

	@property
	def denominator(self) -> Type[Measurement]:
		return getattr(self, '_denominator', None)

	@property
	def numeratorClass(self) -> Type[Measurement]:
		return getattr(self, '_numerator', None)

	@numeratorClass.setter
	def numeratorClass(self, value: Type[Measurement]):
		self._numerator = value

	@property
	def denominatorClass(self) -> Type[Measurement]:
		return getattr(self, '_denominator', None)

	@denominatorClass.setter
	def denominatorClass(self, value: Type[Measurement]):
		self._denominator = value

	@property
	def pluralName(self) -> Optional[str]:
		if not self.numerator.isGeneric:
			return None
		return f'{self.numerator.pluralName}/{self.denominator.name}'

	@property
	def pluralUnit(self) -> Optional[str]:
		if not self.numerator.isGeneric:
			return None
		return f'{self.numerator.pluralUnit}/{self.denominator.unit}'

	@property
	def unitArray(mcs):
		return [mcs.numerator.unit, mcs.denominator.unit]

	@property
	def Generic(self) -> Type:
		genericBases = self.genericBases
		genericBases = sorted(
			(
				base for base in genericBases
				if base is not self
				   and issubclass(base, DerivedMeasurement)
				   and base.numerator.isGeneric
				   and base.denominator.isGeneric),
			key=lambda base: self.__mro__.index(base)
		)
		return next(iter(genericBases), None)

	@property
	def genericBases(self):
		return super().genericBases | self.numerator.genericBases | self.denominator.genericBases

	@property
	def compatibleUnits(cls) -> dict:
		parentCls = cls.type
		n = parentCls.numerator.type.compatibleUnits
		d = parentCls.denominator.type.compatibleUnits
		specifiedUnits = {u: t for t in parentCls.subTypes if (u := getattr(t, '_unit', None))}
		return {**specifiedUnits, **n, **d}

	# @property
	# def subTypes(cls):
	# 	subs = super().subTypes
	# 	fixed = cls.fixedUnits
	# 	if fixed[0] is None:
	# 		subs |= cls.numerator.subTypes
	# 	else:
	# 		subs |= {cls.numerator}
	# 	if fixed[1] is None:
	# 		subs |= cls.denominator.subTypes
	# 	else:
	# 		subs |= {cls.denominator}
	# 	return subs

	@property
	def unit(cls):
		clsUnit = getattr(cls, '_unit', None)
		if clsUnit:
			return clsUnit
		numUnit = cls.numerator.unit or f'({cls.numerator.__name__})'
		denUnit = cls.denominator.unit or f'({cls.denominator.__name__})'
		return f'{numUnit}/{denUnit}'

	def __parse_unit__(self, unit: str | None):
		if unit is None:
			return None
		units = re.sub(r'((?<=[A-Z_\-\s])(per)(?=[A-Z_\-\s]))|\\', '/', unit).split('/')
		return tuple(i.strip(' ') for i in units)

	@property
	def id(cls):
		clsID = getattr(cls, '_id', None)
		if clsID:
			return clsID
		clsUnit = getattr(cls, '_unit', None)
		if clsUnit:
			return clsUnit

		numUnit = cls.numerator.id or f'{cls.numerator.__name__}'
		denUnit = cls.denominator.id or f'{cls.denominator.__name__}'
		return f'{cls.__name__}({numUnit}/{denUnit})'

	@property
	def type(cls):
		if (t := getattr(cls, '_type', None)) is not None:
			return t
		bases = cls.genericBases
		n = {i for i in bases if (j := getattr(i, 'numerator', None)) is not None and j.isGeneric}
		d = {i for i in bases if (j := getattr(i, 'denominator', None)) is not None and j.isGeneric}
		_both = n | d
		return sorted(_both, key=lambda i: cls.mro().index(i))[0]

	@property
	@lru_cache(maxsize=512)
	def __unitDict__(self) -> ChainMap:
		return ChainMap(self.numerator.__unitDict__, self.denominator.__unitDict__)


metric: Final = 'metric'
imperial: Final = 'imperial'
both: Final = 'mixed'
systemName: Final = Literal['metric', 'imperial', 'mixed']


class Dimension(MetaUnitClass):
	__systems__: Dict[str, Type['Dimension']] = {metric: {}, imperial: {}, both: ChainMap()}
	_system: systemName
	__dimensions__: Type['Dimension']

	def __new__(mcs, name, bases, attrs, *args, **kwargs):
		system = kwargs.get('system', None)
		if system is not None:
			attrs['_systemName'] = system

		if symbol := kwargs.pop('symbol', None):
			attrs['_symbol'] = symbol

		newCls = super().__new__(mcs, name, bases, attrs, **kwargs)

		if system is not None:
			newCls._system = newCls
		if dimension := getattr(newCls, 'dimension', None):
			dimension.register(newCls)
		else:
			mcs.registerDimension(newCls)

		return newCls

	@classmethod
	def registerDimension(mcs, dimension):
		systems = mcs.__systems__
		dimension.__dimension__ = dimension
		if dimension not in systems[both]:
			dimensionDictMetric, dimensionDictImperial = {}, {}

			dimensionDictBoth = ChainMap(dimensionDictMetric, dimensionDictImperial)

			systems[metric][dimension] = dimensionDictMetric
			systems[metric][dimension.__name__] = dimensionDictImperial

			systems[imperial][dimension] = dimensionDictImperial
			systems[imperial][dimension.__name__] = dimensionDictImperial

			systems[both][dimension] = dimensionDictBoth
			systems[both][dimension.__name__] = dimensionDictBoth

	@classmethod
	def register(mcs, newCls):
		systems = mcs.__systems__

		dimension = newCls.dimension or newCls

		unit = newCls.unit
		if unit:
			if newCls.systemName is both:
				systems[imperial][dimension][unit] = newCls
				systems[metric][dimension][unit] = newCls
			else:
				systems[newCls.systemName][dimension][unit] = newCls
		else:
			newCls._unitSystem = newCls
			newCls._systemID = f'{newCls.__dimension__.__name__}[{newCls._systemName.title()}]'
		# newCls.__name__ = newCls.systemName

	@property
	def isGeneric(cls) -> bool:
		return cls is cls.__dimension__ or (cls is cls._system and bool(cls.__subclasses__()))

	@property
	def systemName(self) -> metric or imperial:
		return getattr(self, '_systemName', None)

	@property
	def system(self) -> str:
		return getattr(self, '_system', None)

	@property
	def systems(self):
		if self.systemName is None:
			return self.__systems__
		return self.__systems__[self.systemName]


class DerivedMeasurement(Measurement, metaclass=DerivedMeasurementMeta):
	_derived = True
	_type: Type[Measurement] = Measurement
	_numerator: Union[Measurement, Type[Measurement]] = Measurement
	_denominator: Union[Measurement, Type[Measurement]] = Measurement

	def __new__(cls, numerator: Measurement, denominator: Measurement = None, *args, **kwargs):
		if denominator is None:
			if (d := getattr(numerator, 'denominator', None)) is not None:
				denominator = d
				numerator = numerator.numerator
			else:
				denominator = 1
		if not cls.numerator.isGeneric:
			numerator = cls.numerator(numerator)
		if not cls.denominatorClass.isGeneric:
			denominator = cls.denominator(denominator)
		value = float(numerator)/float(denominator)

		if isinstance(numerator, cls.numeratorClass) and isinstance(denominator, cls.denominatorClass):
			if cls.isGeneric:
				cls = cls[type(numerator):type(denominator)]
		elif items := [i for i in cls.type.subTypes if i.denominator is type(denominator)]:
			if len(items) == 1:
				cls = items[0]
			else:
				items = [i for i in items if i.numerator is type(numerator)]
				if len(items) == 1:
					cls = items[0]
				else:
					raise TypeError(f'{cls.__name__} does not support {numerator.__class__.__name__} as numerator')
		elif cls.isGeneric:
			cls = cls[type(numerator):type(denominator)]
		return Measurement.__new__(cls, value, *args, **kwargs)

	def __class_getitem__(cls, item):
		# By default, this method respects fixed units
		# To get a class regardless of fixed units, use cls.getClass(item)
		if isinstance(item, str):
			item = cls.__parse_unit__(item)
		if isinstance(item, tuple):
			item = tuple(cls.type.__findUnitClass__(i) if isinstance(i, str) else i for i in item)
		return cls.__getSlice__(HashSlice(item))

	@classmethod
	def getClass(cls, units):
		if isinstance(units, str):
			units = units,
		clsType = cls.type
		units = tuple(clsType.__findUnitClass__(i) for i in units)
		n = next((i for i in units if cls.numerator.dimension is not None and issubclass(i, cls.numerator.dimension)), None) or cls.numeratorClass
		d = next((i for i in units if cls.denominator.dimension is not None and issubclass(i, cls.denominator.dimension)), None) or cls.denominatorClass
		return cls[n:d:True]

	@classmethod
	@lru_cache(maxsize=32)
	def __getSlice__(cls, item: HashSlice, ignoreFixed=False):
		if isinstance(getattr(item, 'step', None), bool):
			ignoreFixed = item.step
		fixedN, fixedD = cls.fixedUnits if not ignoreFixed else (None, None)
		if item in cls.__dict__:
			return cls.__dict__[item]
		if isinstance(item, HashSlice):
			sliceNType, sliceDType = fixedN or item.start, fixedD or item.stop
			nType, dType = cls.numerator, cls.denominator
			if isinstance(sliceNType, str):
				sliceNType = nType.__findUnitClass__(sliceNType)
			if isinstance(sliceDType, str):
				sliceDType = dType.__findUnitClass__(sliceDType)
			if sliceNType is not None and not isinstance(sliceNType, type):
				sliceNType = type(sliceNType)
			elif sliceNType is None:
				sliceNType = nType

			if sliceDType is not None and not isinstance(sliceDType, type):
				sliceDType = type(sliceDType)
			elif sliceDType is None:
				sliceDType = dType

			genericType = cls if cls.isGeneric else cls.type
			if ignoreFixed and any(i is not None for i in genericType.fixedUnits):
				genericType = cls.Generic
			subTypes = genericType.subTypes
			sameDType = {i for i in subTypes if i.denominator is sliceDType}
			sameNType = {i for i in subTypes if i.numerator is sliceNType}

			if exactMatch := sameDType & sameNType:
				matchedType = exactMatch.pop()
				if not issubclass(matchedType._type, genericType) and issubclass(cls._type, matchedType._type):
					return type(f'{cls._type.__name__}[{matchedType.unit}]', (matchedType, cls._type), {})

			if sameDType and (generics := {i for i in sameDType if i.numeratorClass.isGeneric}):
				if len(generics) == 1:
					generic = generics.pop()
					if not dType.isGeneric:
						return generic[sliceNType or nType]
					if generic.denominatorClass is sliceDType:
						return generic[sliceNType or nType]
			if sameNType and (generics := {i for i in sameNType if i.denominatorClass.isGeneric}):
				if len(generics) == 1:
					generic = generics.pop()
					if not nType.isGeneric:
						return generic[sliceDType or dType]
					if generic.numeratorClass is sliceNType:
						return generic[sliceDType or dType]

			name = f'{genericType.__name__}[{sliceNType.id or sliceNType.__name__}/{sliceDType.id or sliceDType.__name__}]'
			return type(name, (genericType,), {}, numerator=sliceNType, denominator=sliceDType)
		item = item.start
		if issubclass(item, cls.numerator):
			if not cls.denominator.isGeneric:
				name = f'{cls.__name__}'
				newCls = type(name, (cls,), {'_numerator': item})
				return newCls
			print(f'{cls.__name__} does not support {item.__name__} as numerator')

		elif issubclass(item, cls.denominator):
			if not cls.numeratorClass.isGeneric:
				name = f'{cls.__name__}[{item.__name__}]'
				newCls = type(name, (cls,), {'_denominator': item})
				return newCls
			print(f'{cls.__name__} does not support {item.__name__} as denominator')

		raise TypeError(f'{cls.__name__} does not support {item.__name__} as unit')

	def __init__(self, numerator, denominator=None, *args, **kwargs):
		if isinstance(numerator, DerivedMeasurement) and denominator is None:
			self._numerator = numerator.numerator
			self._denominator = numerator.denominator
		elif denominator is None:
			self._numerator = numerator
			self._denominator = self.denominator(1)
		else:
			self._numerator = numerator
			self._denominator = denominator
		if not isinstance(self.numerator, type(self).numerator):
			self._numerator = type(self).numerator(self._numerator)
		if not isinstance(self.denominator, type(self).denominator):
			self._denominator = type(self).denominator(self._denominator)
		Measurement.__init__(self, float(self._numerator)/float(self._denominator), *args, **kwargs)

	def __pow__(self, power, modulo=None):
		return self.__class__(super(Measurement, self).__pow__(power, modulo), self._denominator)

	def __getitem__(self, item):
		if isinstance(item, str):
			item = type(self).__parse_unit__(item)
			if len(item) == 2:
				item = HashSlice(*item, True)
			elif len(item) == 1:
				item = item[0]
				cls = self.type.__findUnitClass__(item)
				if issubclass(cls, self.type):
					return cls(self.numerator, self.denominator)
				if issubclass(cls, self.numerator.type):
					cls = type(self)[cls: type(self.denominator): True]
				else:
					cls = type(self)[type(self.numerator): cls: True]
				return cls(self.numerator, self.denominator)
		if hasattr(item, 'denominator'):
			item = HashSlice(item.numerator, item.denominator, True)

		if isinstance(item, (slice, HashSlice)):
			newCls = self.type[item]
			if newCls is not type(self):
				return newCls(self.numerator, self.denominator)
			numerator = item.start or self._numerator
			denominator = item.stop or self._denominator

			if isinstance(numerator, str):
				numeratorCls = self.type.__findUnitClass__(numerator)
			else:
				numeratorCls = type(self).numerator

			if isinstance(denominator, str):
				denomDict = self._denominator.__unitDict__
				denominatorCls = denomDict.get(denominator, None) or {
					v.__name__.lower(): v for k, v in denomDict.items()
					if v.__name__.lower() == denominator.lower()
				}.get(denominator, None) or type(self).denominator
			else:
				denominatorCls = type(self).denominator

			return self.__class__[numeratorCls, denominatorCls, True](self._numerator, self._denominator)
		return self.__class__(self._numerator[item], self._denominator)

	def transform(self, other, transformTo: type = None):
		if isinstance(other, type) and issubclass(other, Measurement):
			other = other(self)
		nonTransferred = ['_unit', '_suffix', '_scale', '_denominator', '_numerator']
		if self.type != other.type:
			log.warning(f'{self.withUnit} and {other.withUnit} are not identical types, this may cause issues')
		other.__dict__.update({key: value for key, value in self.__dict__.items() if key not in nonTransferred and value is not None})
		if other._updateFunction:
			other._updateFunction(other)
		return other

	def _convert(self, other):
		if isinstance(other, (int, float)):
			return self.__class__(self.n.__class__(other), self.d.__class__(1))
		if isinstance(other, self.n.type):
			try:
				other = other[self.n.unit]
			except AttributeError:
				pass
			return self.__class__(other, self.d)
		else:
			return super(DerivedMeasurement, self)._convert(other)

	# TODO: Implement into child classes
	def _getUnit(self) -> List[str]:
		return config['LocalUnits'][str(self._type)].split('/')

	def _getUnitTypes(self):
		return self._numerator.type, self._denominator.type

	@property
	def type(self) -> Type[Measurement]:
		return type(self).type

	@property
	def localize(self):
		try:
			units = type(self).localizedUnit
			if len(units) == 1:
				unit = units[0]
				if issubclass(unit, self.type):
					return unit(self.numerator, self.denominator)
				if issubclass(unit.type, self.n.type):
					n, d = unit, None
				elif issubclass(unit.type, self.d.type):
					n, d = None, unit
				else:
					n, d = self.n, self.d
			else:
				n, d = units
			t = self.type[n:d:True]
			return t(self.n, self.d)
		except KeyError:
			log.error('Measurement {} was unable to localize from {}'.format(self.name, self.unit))
			return self

	@property
	def unit(self):
		return getattr(self, '_unit', '') or f'{self.numerator.unit}/{self.denominator.unit}'

	@property
	def unitArray(self) -> list[str]:
		return [*self._numerator.unitArray, *self._denominator.unitArray]

	@property
	def numerator(self) -> Measurement:
		return self._numerator

	n = numerator

	@property
	def denominator(self) -> Measurement:
		return self._denominator

	d = denominator

	@property
	def dArr(self):
		arr = []
		if isinstance(self.n, DerivedMeasurement):
			arr += [*self.n.dArr]
		if isinstance(self.d, DerivedMeasurement):
			arr += [*self.d.dArr]
		else:
			arr += [self.d.unit]
		return arr

	@property
	def nArr(self):
		arr = []
		if isinstance(self.d, DerivedMeasurement):
			arr += [*self.n.nArr]
		if isinstance(self.n, DerivedMeasurement):
			arr += [*self.n.nArr]
		else:
			arr += [self.n.unit]
		return arr



from ..length import Length
from ..mass import Mass
from .. import Time
from .. import DerivedMeasurement
from .. import Direction
from .. import Measurement
from .. import ScalingMeasurement

from .rate import DistanceOverTime
from .wind import Wind
from .precipitation import Precipitation
from .volume import Volume
from .density import Density
from .partsPer import PartsPer

__all__ = ['Volume', 'DistanceOverTime', 'Precipitation', 'Density', 'PartsPer', 'Wind']



from typing import Union

from ..base.Decorators import UnitType
from . import Measurement, Length

__all__ = ['Volume']


@UnitType
class Volume(Measurement):
	CubicFoot: type
	CubicMeter: type

	_x: Length
	_y: Length
	_z: Length
	_unitClass = Length
	_cube: bool

	def __init__(self, x: Union[Length, int, float], y: Union[Length, int, float] = None, z: Union[Length, int, float] = None):
		if isinstance(x, Volume):
			y = x.y
			z = x.z
			x = x.x
		if not (y and z):
			x = float(x)
			x **= 1./3.
			y = x
			z = x
		self._x = self._unitClass(x)
		self._y = self._unitClass(y)
		self._z = self._unitClass(z)
		float.__init__(x*y*z)

	def __new__(cls, x: Length, y: Length = None, z: Length = None):
		if isinstance(x, Volume):
			y = x.y
			z = x.z
			x = x.x
		if not (y and z):
			x = float(x)
			x **= 1./3.
			y = x
			z = x
		x = cls._unitClass(x)
		y = cls._unitClass(y)
		z = cls._unitClass(z)
		value = float(x)*float(y)*float(z)
		return float.__new__(cls, value)

	@property
	def width(self) -> Length:
		return self._x

	@property
	def length(self) -> Length:
		return self._y

	@property
	def depth(self) -> Length:
		return self._z

	@property
	def unit(self) -> str:
		return self._x.unit + "³"

	def scale(self, value: float):
		x = self._x * value
		y = self._y * value
		z = self._z * value
		return x, y, z

	@property
	def x(self):
		return self._x

	@property
	def y(self):
		return self._y

	@property
	def z(self):
		return self._z

# class CubicMeter(Volume, Length.Meter, unit='m³', aliases=('m3', 'm^3', 'm'), dimension=(Length, Length, Length)):
# 	_x: Length.Meter
# 	_y: Length.Meter
# 	_z: Length.Meter
# 	_unitClass = Length.Meter
#
# 	@property
# 	def ft(self):
# 		x = self._x.ft
# 		y = self._y.ft
# 		z = self._z.ft
# 		return CubicFoot(x, y, z)
#
# 	@property
# 	def m(self):
# 		return self
#
#
# class CubicFoot(Volume, Length.Foot, unit='ft³', aliases=('ft3', 'ft^3', 'ft'), dimension=(Length, Length, Length)):
# 	_x: Length.Foot
# 	_y: Length.Foot
# 	_z: Length.Foot
# 	_unitClass = Length.Foot
#
# 	@property
# 	def ft(self):
# 		return self
#
# 	@property
# 	def m(self):
# 		x = self._x.m
# 		y = self._y.m
# 		z = self._z.m
# 		return CubicMeter(x, y, z)
#
#
# Volume.CubicMeter = CubicMeter
# Volume.CubicFoot = CubicFoot



__all__ = ['RSSI']

from ..base.Decorators import UnitType
from ..base import Measurement


@UnitType
class RSSI(Measurement):
	_unit = 'dBm'
	_strings = ['Perfect', 'Great', 'Good', 'Bad', 'None']
	_values = [-30, -67, -70, -80, -90]
	_precision = 0
	_max = 2

	@property
	def string(self):
		v = int(self)
		i = 0
		while v < self._values[i] and i < 6:
			i += 1
		return self._strings[i]
